# -*- coding: utf-8 -*-
"""
reV generation module.
"""

import copy
import json
import logging
import os
import pprint

import numpy as np
import pandas as pd
from rex.multi_file_resource import MultiFileResource
from rex.multi_res_resource import MultiResolutionResource
from rex.resource import Resource
from rex.utilities.utilities import check_res_file

from reV.generation.base import BaseGen
from reV.SAM.generation import (
    Geothermal,
    LinearDirectSteam,
    MhkWave,
    PvSamv1,
    PvWattsv5,
    PvWattsv7,
    PvWattsv8,
    SolarWaterHeat,
    TcsMoltenSalt,
    TroughPhysicalHeat,
    WindPower,
)
from reV.utilities import ModuleName, ResourceMetaField, SupplyCurveField
from reV.utilities.exceptions import (
    ConfigError,
    InputError,
    ProjectPointsValueError,
)

logger = logging.getLogger(__name__)


ATTR_DIR = os.path.dirname(os.path.realpath(__file__))
ATTR_DIR = os.path.join(ATTR_DIR, "output_attributes")
with open(os.path.join(ATTR_DIR, "other.json")) as f:
    OTHER_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, "generation.json")) as f:
    GEN_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, "linear_fresnel.json")) as f:
    LIN_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, "solar_water_heat.json")) as f:
    SWH_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, "trough_heat.json")) as f:
    TPPH_ATTRS = json.load(f)


class Gen(BaseGen):
    """Gen"""

    # Mapping of reV technology strings to SAM generation objects
    OPTIONS = {
        "geothermal": Geothermal,
        "lineardirectsteam": LinearDirectSteam,
        "mhkwave": MhkWave,
        "pvsamv1": PvSamv1,
        "pvwattsv5": PvWattsv5,
        "pvwattsv7": PvWattsv7,
        "pvwattsv8": PvWattsv8,
        "solarwaterheat": SolarWaterHeat,
        "tcsmoltensalt": TcsMoltenSalt,
        "troughphysicalheat": TroughPhysicalHeat,
        "windpower": WindPower,
    }

    """reV technology options."""

    # Mapping of reV generation outputs to scale factors and units.
    # Type is scalar or array and corresponds to the SAM single-site output
    OUT_ATTRS = copy.deepcopy(OTHER_ATTRS)
    OUT_ATTRS.update(GEN_ATTRS)
    OUT_ATTRS.update(LIN_ATTRS)
    OUT_ATTRS.update(SWH_ATTRS)
    OUT_ATTRS.update(TPPH_ATTRS)
    OUT_ATTRS.update(BaseGen.ECON_ATTRS)

    def __init__(
        self,
        technology,
        project_points,
        sam_files,
        resource_file,
        low_res_resource_file=None,
        output_request=("cf_mean",),
        site_data=None,
        curtailment=None,
        gid_map=None,
        drop_leap=False,
        sites_per_worker=None,
        memory_utilization_limit=0.4,
        scale_outputs=True,
        write_mapped_gids=False,
        bias_correct=None,
    ):
        """ReV generation analysis class.

        ``reV`` generation analysis runs SAM simulations by piping in
        renewable energy resource data (usually from the NSRDB or WTK),
        loading the SAM config, and then executing the PySAM compute
        module for a given technology. See the documentation for the
        ``reV`` SAM class (e.g. :class:`reV.SAM.generation.WindPower`,
        :class:`reV.SAM.generation.PvWattsv8`,
        :class:`reV.SAM.generation.Geothermal`, etc.) for info on the
        allowed and/or required SAM config file inputs. If economic
        parameters are supplied in the SAM config, then you can bundle a
        "follow-on" econ calculation by just adding the desired econ
        output keys to the `output_request`. You can request ``reV`` to
        run the analysis for one or more "sites", which correspond to
        the meta indices in the resource data (also commonly called the
        ``gid's``).

        Examples
        --------
        The following is an example of the most simple way to run reV
        generation. Note that the ``TESTDATADIR`` refers to the local cloned
        repository and will need to be replaced with a valid path if you
        installed ``reV`` via a simple pip install.

        >>> import os
        >>> from reV import Gen, TESTDATADIR
        >>>
        >>> sam_tech = 'pvwattsv8'
        >>> sites = 0
        >>> fp_sam = os.path.join(TESTDATADIR, 'SAM/naris_pv_1axis_inv13.json')
        >>> fp_res = os.path.join(TESTDATADIR, 'nsrdb/ri_100_nsrdb_2013.h5')
        >>>
        >>> gen = Gen(sam_tech, sites, fp_sam, fp_res)
        >>> gen.run()
        >>>
        >>> gen.out
        {'cf_mean': array([0.16966143], dtype=float32)}
        >>>
        >>> sites = [3, 4, 7, 9]
        >>> req = ('cf_mean', 'lcoe_fcr')
        >>> gen = Gen(sam_tech, sites, fp_sam, fp_res, output_request=req)
        >>> gen.run()
        >>>
        >>> gen.out
        {'fixed_charge_rate': array([0.096, 0.096, 0.096, 0.096],
         'base_capital_cost': array([39767200, 39767200, 39767200, 39767200],
         'base_variable_operating_cost': array([0, 0, 0, 0],
         'base_fixed_operating_cost': array([260000, 260000, 260000, 260000],
         'capital_cost': array([39767200, 39767200, 39767200, 39767200],
         'fixed_operating_cost': array([260000, 260000, 260000, 260000],
         'variable_operating_cost': array([0, 0, 0, 0],
         'capital_cost_multiplier': array([1, 1, 1, 1],
         'cf_mean': array([0.17859147, 0.17869979, 0.1834818 , 0.18646291],
         'lcoe_fcr': array([130.32126, 130.24226, 126.84782, 124.81981]}

        Parameters
        ----------
        technology : str
            String indicating which SAM technology to analyze. Must be
            one of the keys of
            :attr:`~reV.generation.generation.Gen.OPTIONS`. The string
            should be lower-cased with spaces and underscores removed.
        project_points : int | list | tuple | str | dict | pd.DataFrame | slice
            Input specifying which sites to process. A single integer
            representing the generation GID of a site may be specified
            to evaluate reV at a single location. A list or tuple of
            integers (or slice) representing the generation GIDs of
            multiple sites can be specified to evaluate reV at multiple
            specific locations. A string pointing to a project points
            CSV file may also be specified. Typically, the CSV contains
            the following columns:

                - ``gid``: Integer specifying the generation GID of each
                  site.
                - ``config``: Key in the `sam_files` input dictionary
                  (see below) corresponding to the SAM configuration to
                  use for each particular site. This value can also be
                  ``None`` (or left out completely) if you specify only
                  a single SAM configuration file as the `sam_files`
                  input.
                - ``capital_cost_multiplier``: This is an *optional*
                  multiplier input that, if included, will be used to
                  regionally scale the ``capital_cost`` input in the SAM
                  config. If you include this column in your CSV, you
                  *do not* need to specify ``capital_cost``, unless you
                  would like that value to vary regionally and
                  independently of the multiplier (i.e. the multiplier
                  will still be applied on top of the ``capital_cost``
                  input).

            The CSV file may also contain other site-specific inputs by
            including a column named after a config keyword (e.g. a
            column called ``wind_turbine_rotor_diameter`` may be
            included to specify a site-specific turbine diameter for
            each location). Columns that do not correspond to a config
            key may also be included, but they will be ignored. A
            DataFrame following the same guidelines as the CSV input
            (or a dictionary that can be used to initialize such a
            DataFrame) may be used for this input as well.

            .. Note:: By default, the generation GID of each site is
              assumed to match the resource GID to be evaluated for that
              site. However, unique generation GID's can be mapped to
              non-unique resource GID's via the `gid_map` input (see the
              documentation for `gid_map` for more details).

        sam_files : dict | str
            A dictionary mapping SAM input configuration ID(s) to SAM
            configuration(s). Keys are the SAM config ID(s) which
            correspond to the ``config`` column in the project points
            CSV. Values for each key are either a path to a
            corresponding SAM config file or a full dictionary
            of SAM config inputs. For example::

                sam_files = {
                    "default": "/path/to/default/sam.json",
                    "onshore": "/path/to/onshore/sam_config.yaml",
                    "offshore": {
                        "sam_key_1": "sam_value_1",
                        "sam_key_2": "sam_value_2",
                        ...
                    },
                    ...
                }

            This input can also be a string pointing to a single SAM
            config file. In this case, the ``config`` column of the
            CSV points input should be set to ``None`` or left out
            completely. See the documentation for the ``reV`` SAM class
            (e.g. :class:`reV.SAM.generation.WindPower`,
            :class:`reV.SAM.generation.PvWattsv8`,
            :class:`reV.SAM.generation.Geothermal`, etc.) for
            info on the allowed and/or required SAM config file inputs.
        resource_file : str
            Filepath to resource data. This input can be path to a
            single resource HDF5 file or a path including a wildcard
            input like ``/h5_dir/prefix*suffix`` (i.e. if your datasets
            for a single year are spread out over multiple files). In
            all cases, the resource data must be readable by
            :py:class:`rex.resource.Resource`
            or :py:class:`rex.multi_file_resource.MultiFileResource`.
            (i.e. the resource data conform to the
            `rex data format <https://tinyurl.com/3fy7v5kx>`_). This
            means the data file(s) must contain a 1D ``time_index``
            dataset indicating the UTC time of observation, a 1D
            ``meta`` dataset represented by a DataFrame with
            site-specific columns, and 2D resource datasets that match
            the dimensions of (``time_index``, ``meta``). The time index
            must start at 00:00 of January 1st of the year under
            consideration, and its shape must be a multiple of 8760.

            .. Note:: If executing ``reV`` from the command line, this
              input string can contain brackets ``{}`` that will be
              filled in by the `analysis_years` input. Alternatively,
              this input can be a list of explicit files to process. In
              this case, the length of the list must match the length of
              the `analysis_years` input exactly, and the path are
              assumed to align with the `analysis_years` (i.e. the first
              path corresponds to the first analysis year, the second
              path corresponds to the second analysis year, and so on).

            .. Important:: If you are using custom resource data (i.e.
              not NSRDB/WTK/Sup3rCC, etc.), ensure the following:

                  - The data conforms to the
                    `rex data format <https://tinyurl.com/3fy7v5kx>`_.
                  - The ``meta`` DataFrame is organized such that every
                    row is a pixel and at least the columns
                    ``latitude``, ``longitude``, ``timezone``, and
                    ``elevation`` are given for each location.
                  - The time index and associated temporal data is in
                    UTC.
                  - The latitude is between -90 and 90 and longitude is
                    between -180 and 180.
                  - For solar data, ensure the DNI/DHI are not zero. You
                    can calculate one of these these inputs from the
                    other using the relationship

                    .. math:: GHI = DNI * cos(SZA) + DHI

        low_res_resource_file : str, optional
            Optional low resolution resource file that will be
            dynamically mapped+interpolated to the nominal-resolution
            `resource_file`. This needs to be of the same format as
            `resource_file` - both files need to be handled by the
            same ``rex Resource`` handler (e.g. ``WindResource``). All
            of the requirements from the `resource_file` apply to this
            input as well. If ``None``, no dynamic mapping to higher
            resolutions is performed. By default, ``None``.
        output_request : list | tuple, optional
            List of output variables requested from SAM. Can be any
            of the parameters in the "Outputs" group of the PySAM module
            (e.g. :py:class:`PySAM.Windpower.Windpower.Outputs`,
            :py:class:`PySAM.Pvwattsv8.Pvwattsv8.Outputs`,
            :py:class:`PySAM.Geothermal.Geothermal.Outputs`, etc.) being
            executed. This list can also include a select number of SAM
            config/resource parameters to include in the output:
            any key in any of the
            `output attribute JSON files <https://tinyurl.com/4bmrpe3j/>`_
            may be requested. If ``cf_mean`` is not included in this
            list, it will automatically be added. Time-series profiles
            requested via this input are output in UTC.

            .. Note:: If you are performing ``reV`` solar runs using
              ``PVWatts`` and would like ``reV`` to include AC capacity
              values in your aggregation/supply curves, then you must
              include the ``"dc_ac_ratio"`` time series as an output in
              `output_request` when running ``reV`` generation. The AC
              capacity outputs will automatically be added during the
              aggregation/supply curve step if the ``"dc_ac_ratio"``
              dataset is detected in the generation file.

            By default, ``('cf_mean',)``.
        site_data : str | pd.DataFrame, optional
            Site-specific input data for SAM calculation. If this input
            is a string, it should be a path that points to a CSV file.
            Otherwise, this input should be a DataFrame with
            pre-extracted site data. Rows in this table should match
            the input sites via a ``gid`` column. The rest of the
            columns should match configuration input keys that will take
            site-specific values. Note that some or all site-specific
            inputs can be specified via the `project_points` input
            table instead. If ``None``, no site-specific data is
            considered.

            .. Note:: This input is often used to provide site-based
               regional capital cost multipliers. ``reV`` does not
               ingest multipliers directly; instead, this file is
               expected to have a ``capital_cost`` column that gives the
               multiplier-adjusted capital cost value for each location.
               Therefore, you *must* re-create this input file every
               time you change your base capital cost assumption.

            By default, ``None``.
        curtailment : dict | str, optional
            Inputs for curtailment parameters, which can be:

                - Explicit namespace of curtailment variables (dict)
                - Pointer to curtailment config file with path (str)

            The allowed key-value input pairs in the curtailment
            configuration are documented as properties of the
            :class:`reV.config.curtailment.Curtailment` class. If
            ``None``, no curtailment is modeled. By default, ``None``.
        gid_map : dict | str, optional
            Mapping of unique integer generation gids (keys) to single
            integer resource gids (values). This enables unique
            generation gids in the project points to map to non-unique
            resource gids, which can be useful when evaluating multiple
            resource datasets in ``reV`` (e.g., forecasted ECMWF
            resource data to complement historical WTK meteorology).
            This input can be a pre-extracted dictionary or a path to a
            JSON or CSV file. If this input points to a CSV file, the
            file must have the columns ``gid`` (which matches the
            project points) and ``gid_map`` (gids to extract from the
            resource input). If ``None``, the GID values in the project
            points are assumed to match the resource GID values.
            By default, ``None``.
        drop_leap : bool, optional
            Drop leap day instead of final day of year when handling
            leap years. By default, ``False``.
        sites_per_worker : int, optional
            Number of sites to run in series on a worker. ``None``
            defaults to the resource file chunk size.
            By default, ``None``.
        memory_utilization_limit : float, optional
            Memory utilization limit (fractional). Must be a value
            between 0 and 1. This input sets how many site results will
            be stored in-memory at any given time before flushing to
            disk. By default, ``0.4``.
        scale_outputs : bool, optional
            Flag to scale outputs in-place immediately upon ``Gen``
            returning data. By default, ``True``.
        write_mapped_gids : bool, optional
            Option to write mapped gids to output meta instead of
            resource gids. By default, ``False``.
        bias_correct : str | pd.DataFrame, optional
            Optional DataFrame or CSV filepath to a wind or solar
            resource bias correction table. This has columns:

                - ``gid``: GID of site (can be index name of dataframe)
                - ``method``: function name from ``rex.bias_correction`` module

            The ``gid`` field should match the true resource ``gid`` regardless
            of the optional ``gid_map`` input. Only ``windspeed`` **or**
            ``GHI`` + ``DNI`` + ``DHI`` are corrected, depending on the
            technology (wind for the former, PV or CSP for the latter). See the
            functions in the ``rex.bias_correction`` module for available
            inputs for ``method``. Any additional kwargs required for the
            requested ``method`` can be input as additional columns in the
            ``bias_correct`` table e.g., for linear bias correction functions
            you can include ``scalar`` and ``adder`` inputs as columns in the
            ``bias_correct`` table on a site-by-site basis. If ``None``, no
            corrections are applied. By default, ``None``.
        """
        pc = self.get_pc(
            points=project_points,
            points_range=None,
            sam_configs=sam_files,
            tech=technology,
            sites_per_worker=sites_per_worker,
            res_file=resource_file,
            curtailment=curtailment,
        )

        super().__init__(
            pc,
            output_request,
            site_data=site_data,
            drop_leap=drop_leap,
            memory_utilization_limit=memory_utilization_limit,
            scale_outputs=scale_outputs,
        )

        if self.tech not in self.OPTIONS:
            msg = (
                'Requested technology "{}" is not available. '
                "reV generation can analyze the following "
                "SAM technologies: {}".format(
                    self.tech, list(self.OPTIONS.keys())
                )
            )
            logger.error(msg)
            raise KeyError(msg)

        self.write_mapped_gids = write_mapped_gids
        self._res_file = resource_file
        self._lr_res_file = low_res_resource_file
        self._sam_module = self.OPTIONS[self.tech]
        self._run_attrs["sam_module"] = self._sam_module.MODULE
        self._run_attrs["res_file"] = resource_file

        self._multi_h5_res, self._hsds = check_res_file(resource_file)
        self._gid_map = self._parse_gid_map(gid_map)
        self._nn_map = self._parse_nn_map()
        self._bc = self._parse_bc(bias_correct)

    @property
    def res_file(self):
        """Get the resource filename and path.

        Returns
        -------
        res_file : str
            Filepath to single resource file, multi-h5 directory,
            or /h5_dir/prefix*suffix
        """
        return self._res_file

    @property
    def lr_res_file(self):
        """Get the (optional) low-resolution resource filename and path.

        Returns
        -------
        str | None
        """
        return self._lr_res_file

    @property
    def meta(self):
        """Get resource meta for all sites in project points.

        Returns
        -------
        meta : pd.DataFrame
            Meta data df for sites in project points. Column names are meta
            data variables, rows are different sites. The row index
            does not indicate the site number if the project points are
            non-sequential or do not start from 0, so a `SiteDataField.GID`
            column is added.
        """
        if self._meta is None:
            res_cls = Resource
            kwargs = {"hsds": self._hsds}
            if self._multi_h5_res:
                res_cls = MultiFileResource
                kwargs = {}

            res_gids = self.project_points.sites
            if self._gid_map is not None:
                res_gids = [self._gid_map[i] for i in res_gids]

            with res_cls(self.res_file, **kwargs) as res:
                meta_len = res.shapes["meta"][0]

                if np.max(res_gids) > meta_len:
                    msg = (
                        "ProjectPoints has a max site gid of {} which is "
                        "out of bounds for the meta data of len {} from "
                        "resource file: {}".format(
                            np.max(res_gids), meta_len, self.res_file
                        )
                    )
                    logger.error(msg)
                    raise ProjectPointsValueError(msg)

                self._meta = res["meta", res_gids]

            self._meta.loc[:, ResourceMetaField.GID] = res_gids
            if self.write_mapped_gids:
                sites = self.project_points.sites
                self._meta.loc[:, ResourceMetaField.GID] = sites
            self._meta.index = self.project_points.sites
            self._meta.index.name = ResourceMetaField.GID
            self._meta.loc[:, "reV_tech"] = self.project_points.tech

        return self._meta

    @property
    def time_index(self):
        """Get the generation resource time index data.

        Returns
        -------
        _time_index : pandas.DatetimeIndex
            Time-series datetime index
        """
        if self._time_index is None:
            if not self._multi_h5_res:
                res_cls = Resource
                kwargs = {"hsds": self._hsds}
            else:
                res_cls = MultiFileResource
                kwargs = {}

            with res_cls(self.res_file, **kwargs) as res:
                time_index = res.time_index

            downscale = self.project_points.sam_config_obj.downscale
            step = self.project_points.sam_config_obj.time_index_step
            if downscale is not None:
                from rex.utilities.downscale import make_time_index

                year = time_index.year[0]
                ds_freq = downscale["frequency"]
                time_index = make_time_index(year, ds_freq)
                logger.info(
                    "reV solar generation running with temporal "
                    'downscaling frequency "{}" with final '
                    "time_index length {}".format(ds_freq, len(time_index))
                )
            elif step is not None:
                time_index = time_index[::step]

            time_index = self.handle_lifetime_index(time_index)
            time_index = self.handle_leap_ti(
                time_index, drop_leap=self._drop_leap
            )

            self._time_index = time_index

        return self._time_index

    def handle_lifetime_index(self, ti):
        """Adjust the time index if modeling full system lifetime.

        Parameters
        ----------
        ti : pandas.DatetimeIndex
            Time-series datetime index with leap days.

        Returns
        -------
        ti : pandas.DatetimeIndex
            Time-series datetime index.
        """
        life_var = "system_use_lifetime_output"
        lifetime_periods = []
        for sam_meta in self.sam_metas.values():
            if life_var in sam_meta and sam_meta[life_var] == 1:
                lifetime_period = sam_meta["analysis_period"]
                lifetime_periods.append(lifetime_period)
            else:
                lifetime_periods.append(1)

        if not any(ltp > 1 for ltp in lifetime_periods):
            return ti

        # Only one time index may be passed, check that lifetime periods match
        n_unique_periods = len(np.unique(lifetime_periods))
        if n_unique_periods != 1:
            msg = (
                "reV cannot handle multiple analysis_periods when "
                "modeling with `system_use_lifetime_output` set "
                "to 1. Found {} different analysis_periods in the SAM "
                "configs".format(n_unique_periods)
            )
            logger.error(msg)
            raise ConfigError(msg)

        # Collect requested variables to check for lifetime compatibility
        array_vars = [
            var for var, attrs in GEN_ATTRS.items() if attrs["type"] == "array"
        ]
        valid_vars = ["gen_profile", "cf_profile", "cf_profile_ac"]
        invalid_vars = set(array_vars) - set(valid_vars)
        invalid_requests = [
            var for var in self.output_request if var in invalid_vars
        ]

        if invalid_requests:
            # SAM does not output full lifetime for all array variables
            msg = (
                "reV can only handle the following output arrays "
                "when modeling with `system_use_lifetime_output` set "
                "to 1: {}. Try running without {}.".format(
                    ", ".join(valid_vars), ", ".join(invalid_requests)
                )
            )
            logger.error(msg)
            raise ConfigError(msg)

        sam_meta = self.sam_metas[next(iter(self.sam_metas))]
        analysis_period = sam_meta["analysis_period"]
        logger.info(
            "reV generation running with a full system "
            "life of {} years.".format(analysis_period)
        )

        old_end = ti[-1]
        new_end = old_end + pd.DateOffset(years=analysis_period - 1)
        step = old_end - ti[-2]
        time_extension = pd.date_range(old_end, new_end, freq=step)
        ti = time_extension.union(ti)

        return ti

    @classmethod
    def _run_single_worker(
        cls,
        points_control,
        tech=None,
        res_file=None,
        lr_res_file=None,
        output_request=None,
        scale_outputs=True,
        gid_map=None,
        nn_map=None,
        bias_correct=None,
    ):
        """Run a SAM generation analysis based on the points_control iterator.

        Parameters
        ----------
        points_control : reV.config.PointsControl
            A PointsControl instance dictating what sites and configs are run.
        tech : str
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed.
        res_file : str
            Filepath to single resource file, multi-h5 directory,
            or /h5_dir/prefix*suffix
        lr_res_file : str | None
            Optional low resolution resource file that will be dynamically
            mapped+interpolated to the nominal-resolution res_file. This
            needs to be of the same format as resource_file, e.g. they both
            need to be handled by the same rex Resource handler such as
            WindResource
        output_request : list | tuple
            Output variables requested from SAM.
        scale_outputs : bool
            Flag to scale outputs in-place immediately upon Gen returning data.
        gid_map : None | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This enables the user to input unique
            generation gids in the project points that map to non-unique
            resource gids. This can be None or a pre-extracted dict.
        nn_map : np.ndarray
            Optional 1D array of nearest neighbor mappings associated with the
            res_file to lr_res_file spatial mapping. For details on this
            argument, see the rex.MultiResolutionResource docstring.
        bias_correct : None | pd.DataFrame
            Optional DataFrame or CSV filepath to a wind or solar
            resource bias correction table. This has columns:

                - ``gid``: GID of site (can be index name of dataframe)
                - ``method``: function name from ``rex.bias_correction`` module

            The ``gid`` field should match the true resource ``gid`` regardless
            of the optional ``gid_map`` input. Only ``windspeed`` **or**
            ``GHI`` + ``DNI`` + ``DHI`` are corrected, depending on the
            technology (wind for the former, PV or CSP for the latter). See the
            functions in the ``rex.bias_correction`` module for available
            inputs for ``method``. Any additional kwargs required for the
            requested ``method`` can be input as additional columns in the
            ``bias_correct`` table e.g., for linear bias correction functions
            you can include ``scalar`` and ``adder`` inputs as columns in the
            ``bias_correct`` table on a site-by-site basis. If ``None``, no
            corrections are applied. By default, ``None``.

        Returns
        -------
        out : dict
            Output dictionary from the SAM reV_run function. Data is scaled
            within this function to the datatype specified in Gen.OUT_ATTRS.
        """

        # Extract the site df from the project points df.
        site_df = points_control.project_points.df
        site_df = site_df.set_index(ResourceMetaField.GID, drop=True)

        # run generation method for specified technology
        try:
            out = cls.OPTIONS[tech].reV_run(
                points_control,
                res_file,
                site_df,
                lr_res_file=lr_res_file,
                output_request=output_request,
                gid_map=gid_map,
                nn_map=nn_map,
                bias_correct=bias_correct,
            )

        except Exception as e:
            out = {}
            logger.exception("Worker failed for PC: {}".format(points_control))
            raise e

        if scale_outputs:
            # dtype convert in-place so no float data is stored unnecessarily
            for site, site_output in out.items():
                for k in site_output.keys():
                    # iterate through variable names in each site's output dict
                    if k in cls.OUT_ATTRS:
                        if out[site][k] is None:
                            continue
                        # get dtype and scale for output variable name
                        dtype = cls.OUT_ATTRS[k].get("dtype", "float32")
                        scale_factor = cls.OUT_ATTRS[k].get("scale_factor", 1)

                        # apply scale factor and dtype
                        out[site][k] *= scale_factor

                        if np.issubdtype(dtype, np.integer):
                            # round after scaling if integer dtype
                            out[site][k] = np.round(out[site][k])

                        if isinstance(out[site][k], np.ndarray):
                            # simple astype for arrays
                            out[site][k] = out[site][k].astype(dtype)
                        else:
                            # use numpy array conversion for scalar values
                            out[site][k] = np.array(
                                [out[site][k]], dtype=dtype
                            )[0]

        return out

    def _parse_gid_map(self, gid_map):
        """
        Parameters
        ----------
        gid_map : None | dict | str
            This can be None, a pre-extracted dict, or a filepath to json or
            csv. If this is a csv, it must have the columns "gid" (which
            matches the project points) and "gid_map" (gids to extract from the
            resource input)

        Returns
        -------
        gid_map : None | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This enables the user to input unique
            generation gids in the project points that map to non-unique
            resource gids.
        """

        if isinstance(gid_map, str):
            if gid_map.endswith(".csv"):
                gid_map = pd.read_csv(gid_map).to_dict()
                msg = f"Need {ResourceMetaField.GID} in gid_map column"
                assert ResourceMetaField.GID in gid_map, msg
                assert "gid_map" in gid_map, 'Need "gid_map" in gid_map column'
                gid_map = {
                    gid_map[ResourceMetaField.GID][i]: gid_map["gid_map"][i]
                    for i in gid_map[ResourceMetaField.GID].keys()
                }

            elif gid_map.endswith(".json"):
                with open(gid_map) as f:
                    gid_map = json.load(f)

        if isinstance(gid_map, dict):
            if not self._multi_h5_res:
                res_cls = Resource
                kwargs = {"hsds": self._hsds}
            else:
                res_cls = MultiFileResource
                kwargs = {}

            with res_cls(self.res_file, **kwargs) as res:
                for gen_gid, res_gid in gid_map.items():
                    msg1 = (
                        "gid_map values must all be int but received "
                        "{}: {}".format(gen_gid, res_gid)
                    )
                    msg2 = (
                        "Could not find the gen_gid to res_gid mapping "
                        "{}: {} in the resource meta data.".format(
                            gen_gid, res_gid
                        )
                    )
                    assert isinstance(gen_gid, int), msg1
                    assert isinstance(res_gid, int), msg1
                    assert res_gid in res.meta.index.values, msg2

                for gen_gid in self.project_points.sites:
                    msg3 = (
                        "Could not find the project points gid {} in the "
                        "gen_gid input of the gid_map.".format(gen_gid)
                    )
                    assert gen_gid in gid_map, msg3

        elif gid_map is not None:
            msg = (
                "Could not parse gid_map, must be None, dict, or path to "
                "csv or json, but received: {}".format(gid_map)
            )
            logger.error(msg)
            raise InputError(msg)

        return gid_map

    def _parse_nn_map(self):
        """Parse a nearest-neighbor spatial mapping array if lr_res_file is
        provided (resource data is at two resolutions and the low-resolution
        data must be mapped to the nominal-resolution data)

        Returns
        -------
        nn_map : np.ndarray
            Optional 1D array of nearest neighbor mappings associated with the
            res_file to lr_res_file spatial mapping. For details on this
            argument, see the rex.MultiResolutionResource docstring.
        """
        nn_map = None
        if self.lr_res_file is not None:
            handler_class = Resource
            if "*" in self.res_file or "*" in self.lr_res_file:
                handler_class = MultiFileResource

            with handler_class(self.res_file) as hr_res, handler_class(
                self.lr_res_file
            ) as lr_res:
                logger.info(
                    "Making nearest neighbor map for multi "
                    "resolution resource data..."
                )
                nn_d, nn_map = MultiResolutionResource.make_nn_map(
                    hr_res, lr_res
                )
                logger.info(
                    "Done making nearest neighbor map for multi "
                    "resolution resource data!"
                )

            logger.info(
                "Made nearest neighbor mapping between nominal-"
                "resolution and low-resolution resource files. "
                "Min / mean / max dist: {:.3f} / {:.3f} / {:.3f}".format(
                    nn_d.min(), nn_d.mean(), nn_d.max()
                )
            )

        return nn_map

    @staticmethod
    def _parse_bc(bias_correct):
        """Parse the bias correction data.

        Parameters
        ----------
        bias_correct : str | pd.DataFrame, optional
            Optional DataFrame or CSV filepath to a wind or solar
            resource bias correction table. This has columns:

                - ``gid``: GID of site (can be index name of dataframe)
                - ``method``: function name from ``rex.bias_correction`` module

            The ``gid`` field should match the true resource ``gid`` regardless
            of the optional ``gid_map`` input. Only ``windspeed`` **or**
            ``GHI`` + ``DNI`` + ``DHI`` are corrected, depending on the
            technology (wind for the former, PV or CSP for the latter). See the
            functions in the ``rex.bias_correction`` module for available
            inputs for ``method``. Any additional kwargs required for the
            requested ``method`` can be input as additional columns in the
            ``bias_correct`` table e.g., for linear bias correction functions
            you can include ``scalar`` and ``adder`` inputs as columns in the
            ``bias_correct`` table on a site-by-site basis. If ``None``, no
            corrections are applied. By default, ``None``.

        Returns
        -------
        bias_correct : None | pd.DataFrame
            Optional DataFrame or CSV filepath to a wind or solar
            resource bias correction table. This has columns:

                - ``gid``: GID of site (can be index name of dataframe)
                - ``method``: function name from ``rex.bias_correction`` module

            The ``gid`` field should match the true resource ``gid`` regardless
            of the optional ``gid_map`` input. Only ``windspeed`` **or**
            ``GHI`` + ``DNI`` + ``DHI`` are corrected, depending on the
            technology (wind for the former, PV or CSP for the latter). See the
            functions in the ``rex.bias_correction`` module for available
            inputs for ``method``. Any additional kwargs required for the
            requested ``method`` can be input as additional columns in the
            ``bias_correct`` table e.g., for linear bias correction functions
            you can include ``scalar`` and ``adder`` inputs as columns in the
            ``bias_correct`` table on a site-by-site basis. If ``None``, no
            corrections are applied. By default, ``None``.
        """

        if isinstance(bias_correct, type(None)):
            return bias_correct

        if isinstance(bias_correct, str):
            bias_correct = pd.read_csv(bias_correct).rename(
                SupplyCurveField.map_to(ResourceMetaField), axis=1
            )

        msg = (
            "Bias correction data must be a filepath to csv or a dataframe "
            "but received: {}".format(type(bias_correct))
        )
        assert isinstance(bias_correct, pd.DataFrame), msg

        msg = (
            "Bias correction table must have {!r} column but only found: "
            "{}".format(ResourceMetaField.GID, list(bias_correct.columns))
        )
        assert (
            ResourceMetaField.GID in bias_correct
            or bias_correct.index.name == ResourceMetaField.GID
        ), msg

        if bias_correct.index.name != ResourceMetaField.GID:
            bias_correct = bias_correct.set_index(ResourceMetaField.GID)

        msg = (
            'Bias correction table must have "method" column but only '
            "found: {}".format(list(bias_correct.columns))
        )
        assert "method" in bias_correct, msg

        return bias_correct

    def _parse_output_request(self, req):
        """Set the output variables requested from generation.

        Parameters
        ----------
        req : list | tuple
            Output variables requested from SAM.

        Returns
        -------
        output_request : list
            Output variables requested from SAM.
        """

        output_request = super()._parse_output_request(req)

        # ensure that cf_mean is requested from output
        if "cf_mean" not in output_request:
            output_request.append("cf_mean")

        if _is_solar_run_with_ac_outputs(self.tech):
            if "dc_ac_ratio" not in output_request:
                output_request.append("dc_ac_ratio")
            for dset in ["cf_mean", "cf_profile"]:
                ac_dset = f"{dset}_ac"
                if dset in output_request and ac_dset not in output_request:
                    output_request.append(ac_dset)

        for request in output_request:
            if request not in self.OUT_ATTRS:
                msg = (
                    'User output request "{}" not recognized. '
                    "Will attempt to extract from PySAM.".format(request)
                )
                logger.debug(msg)

        return list(set(output_request))

    def _reduce_kwargs(self, pc, **kwargs):
        """Reduce the global kwargs on a per-worker basis to reduce memory
        footprint

        Parameters
        ----------
        pc : PointsControl
            PointsControl object for a single worker chunk
        kwargs : dict
            reV generation kwargs for all gids that needs to be reduced before
            being sent to ``_run_single_worker()``

        Returns
        -------
        kwargs : dict
            Same as input but reduced just for the gids in pc
        """

        gids = pc.project_points.gids
        gid_map = kwargs.get("gid_map", None)
        bias_correct = kwargs.get("bias_correct", None)

        if bias_correct is not None:
            if gid_map is not None:
                gids = [gid_map[gid] for gid in gids]

            mask = bias_correct.index.isin(gids)
            kwargs["bias_correct"] = bias_correct[mask]

        return kwargs

    def run(self, out_fpath=None, max_workers=1, timeout=1800, pool_size=None):
        """Execute a parallel reV generation run with smart data flushing.

        Parameters
        ----------
        out_fpath : str, optional
            Path to output file. If ``None``, no output file will
            be written. If the filepath is specified but the module name
            (generation) and/or resource data year is not included, the
            module name and/or resource data year will get added to the
            output file name. By default, ``None``.
        max_workers : int, optional
            Number of local workers to run on. If ``None``, or if
            running from the command line and omitting this argument
            from your config file completely, this input is set to
            ``os.cpu_count()``. Otherwise, the default is ``1``.
        timeout : int, optional
            Number of seconds to wait for parallel run iteration to
            complete before returning zeros. By default, ``1800``
            seconds.
        pool_size : int, optional
            Number of futures to submit to a single process pool for
            parallel futures. If ``None``, the pool size is set to
            ``os.cpu_count() * 2``. By default, ``None``.

        Returns
        -------
        str | None
            Path to output HDF5 file, or ``None`` if results were not
            written to disk.
        """
        # initialize output file
        self._init_fpath(out_fpath, module=ModuleName.GENERATION)
        self._init_h5()
        self._init_out_arrays()
        if pool_size is None:
            pool_size = os.cpu_count() * 2

        kwargs = {
            "tech": self.tech,
            "res_file": self.res_file,
            "lr_res_file": self.lr_res_file,
            "output_request": self.output_request,
            "scale_outputs": self.scale_outputs,
            "gid_map": self._gid_map,
            "nn_map": self._nn_map,
            "bias_correct": self._bc,
        }

        logger.info(
            "Running reV generation for: {}".format(self.points_control)
        )
        logger.debug(
            'The following project points were specified: "{}"'.format(
                self.project_points
            )
        )
        logger.debug(
            "The following SAM configs are available to this run:\n{}".format(
                pprint.pformat(self.sam_configs, indent=4)
            )
        )
        logger.debug(
            "The SAM output variables have been requested:\n{}".format(
                self.output_request
            )
        )

        # use serial or parallel execution control based on max_workers
        try:
            if max_workers == 1:
                logger.debug(
                    "Running serial generation for: {}".format(
                        self.points_control
                    )
                )
                for i, pc_sub in enumerate(self.points_control):
                    self.out = self._run_single_worker(pc_sub, **kwargs)
                    logger.info(
                        "Finished reV gen serial compute for: {} "
                        "(iteration {} out of {})".format(
                            pc_sub, i + 1, len(self.points_control)
                        )
                    )
                self.flush()
            else:
                logger.debug(
                    "Running parallel generation for: {}".format(
                        self.points_control
                    )
                )
                self._parallel_run(
                    max_workers=max_workers,
                    pool_size=pool_size,
                    timeout=timeout,
                    **kwargs,
                )

        except Exception as e:
            logger.exception("reV generation failed!")
            raise e

        return self._out_fpath


def _is_solar_run_with_ac_outputs(tech):
    """True if tech is pvwattsv8+"""
    if "pvwatts" not in tech.casefold():
        return False
    return tech.casefold() not in {f"pvwattsv{i}" for i in range(8)}



"""
setup.py
"""
import os
from codecs import open
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from subprocess import check_call
import shlex
from warnings import warn

here = os.path.abspath(os.path.dirname(__file__))

PYPI_DISALLOWED_RST = {"raw::", "<p", "</p", "<img", "---------"}
readme = []
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    for line in f.readlines():
        if any(substr in line for substr in PYPI_DISALLOWED_RST):
            continue
        readme.append(line)

readme = "".join(readme).replace(
    "A visual summary of this process is given below:", ""
)

with open(os.path.join(here, "reV", "version.py"), encoding="utf-8") as f:
    version = f.read()

version = version.split('=')[-1].strip().strip('"').strip("'")


class PostDevelopCommand(develop):
    """
    Class to run post setup commands
    """

    def run(self):
        """
        Run method that tries to install pre-commit hooks
        """
        try:
            check_call(shlex.split("pre-commit install"))
        except Exception as e:
            warn("Unable to run 'pre-commit install': {}"
                 .format(e))

        develop.run(self)


with open("requirements.txt") as f:
    install_requires = f.readlines()

test_requires = ["pytest>=5.2", ]
description = ("National Renewable Energy Laboratory's (NREL's) Renewable "
               "Energy Potential(V) Model: reV")

setup(
    name="NREL-reV",
    version=version,
    description=description,
    long_description=readme,
    author="Galen Maclaurin",
    author_email="galen.maclaurin@nrel.gov",
    url="https://nrel.github.io/reV/",
    packages=find_packages(),
    package_dir={"reV": "reV"},
    entry_points={
        "console_scripts": ["reV=reV.cli:main",
                            "reV-bespoke=reV.bespoke.cli_bespoke:main",
                            "reV-collect=reV.handlers.cli_collect:main",
                            "reV-econ=reV.econ.cli_econ:main",
                            "reV-gen=reV.generation.cli_gen:main",
                            "reV-multiyear=reV.handlers.cli_multi_year:main",
                            ("reV-supply-curve-aggregation=reV.supply_curve."
                             "cli_sc_aggregation:main"),
                            ("reV-supply-curve=reV.supply_curve."
                             "cli_supply_curve:main"),
                            "reV-nrwal=reV.nrwal.cli_nrwal:main",
                            ("reV-rep-profiles=reV.rep_profiles."
                             "cli_rep_profiles:main"),
                            "reV-QA-QC=reV.qa_qc.cli_qa_qc:main",
                            "reV-hybrids=reV.hybrids.cli_hybrids:main",
                            ("reV-project-points=reV.config."
                             "cli_project_points:project_points"),
                            ],
    },
    package_data={'reV': ['SAM/defaults/*.json', 'SAM/defaults/*.csv',
                          'generation/output_attributes/*.json']},
    include_package_data=True,
    license="BSD 3-Clause",
    zip_safe=False,
    keywords="reV",
    python_requires='>=3.8',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    test_suite="tests",
    install_requires=install_requires,
    extras_require={
        "test": test_requires,
        "dev": test_requires + ["flake8", "pre-commit", "pylint"],
    },
    cmdclass={"develop": PostDevelopCommand},
)



# -*- coding: utf-8 -*-
"""reV econ module (lcoe-fcr, single owner, etc...)"""
import logging
import os
import pprint
from warnings import warn

import numpy as np
import pandas as pd
from rex.multi_file_resource import MultiFileResource
from rex.resource import Resource
from rex.utilities.utilities import check_res_file

from reV.config.project_points import PointsControl
from reV.generation.base import BaseGen
from reV.handlers.outputs import Outputs
from reV.SAM.econ import LCOE as SAM_LCOE
from reV.SAM.econ import SingleOwner
from reV.SAM.windbos import WindBos
from reV.utilities import ModuleName, ResourceMetaField
from reV.utilities.exceptions import ExecutionError, OffshoreWindInputWarning

logger = logging.getLogger(__name__)


class Econ(BaseGen):
    """Econ"""

    # Mapping of reV econ output strings to SAM econ modules
    OPTIONS = {
        "lcoe_fcr": SAM_LCOE,
        "ppa_price": SingleOwner,
        "project_return_aftertax_npv": SingleOwner,
        "lcoe_real": SingleOwner,
        "lcoe_nom": SingleOwner,
        "flip_actual_irr": SingleOwner,
        "gross_revenue": SingleOwner,
        "total_installed_cost": WindBos,
        "turbine_cost": WindBos,
        "sales_tax_cost": WindBos,
        "bos_cost": WindBos,
        "fixed_charge_rate": SAM_LCOE,
        "capital_cost": SAM_LCOE,
        "fixed_operating_cost": SAM_LCOE,
        "variable_operating_cost": SAM_LCOE,
    }
    """Available ``reV`` econ `output_request` options"""

    # Mapping of reV econ outputs to scale factors and units.
    # Type is scalar or array and corresponds to the SAM single-site output
    OUT_ATTRS = BaseGen.ECON_ATTRS

    def __init__(self, project_points, sam_files, cf_file, site_data=None,
                 output_request=('lcoe_fcr',), sites_per_worker=100,
                 memory_utilization_limit=0.4, append=False):
        """ReV econ analysis class.

        ``reV`` econ analysis runs SAM econ calculations, typically to
        compute LCOE (using :py:class:`PySAM.Lcoefcr.Lcoefcr`), though
        :py:class:`PySAM.Singleowner.Singleowner` or
        :py:class:`PySAM.Windbos.Windbos` calculations can also be
        performed simply by requesting outputs from those computation
        modules. See the keys of
        :attr:`Econ.OPTIONS <reV.econ.econ.Econ.OPTIONS>` for all
        available econ outputs. Econ computations rely on an input a
        generation (i.e. capacity factor) profile. You can request
        ``reV`` to run the analysis for one or more "sites", which
        correspond to the meta indices in the generation data.

        Parameters
        ----------
        project_points : int | list | tuple | str | dict | pd.DataFrame | slice
            Input specifying which sites to process. A single integer
            representing the GID of a site may be specified to evaluate
            reV at a single location. A list or tuple of integers
            (or slice) representing the GIDs of multiple sites can be
            specified to evaluate reV at multiple specific locations.
            A string pointing to a project points CSV file may also be
            specified. Typically, the CSV contains the following
            columns:

                - ``gid``: Integer specifying the generation GID of each
                  site.
                - ``config``: Key in the `sam_files` input dictionary
                  (see below) corresponding to the SAM configuration to
                  use for each particular site. This value can also be
                  ``None`` (or left out completely) if you specify only
                  a single SAM configuration file as the `sam_files`
                  input.
                - ``capital_cost_multiplier``: This is an *optional*
                  multiplier input that, if included, will be used to
                  regionally scale the ``capital_cost`` input in the SAM
                  config. If you include this column in your CSV, you
                  *do not* need to specify ``capital_cost``, unless you
                  would like that value to vary regionally and
                  independently of the multiplier (i.e. the multiplier
                  will still be applied on top of the ``capital_cost``
                  input).

            The CSV file may also contain other site-specific inputs by
            including a column named after a config keyword (e.g. a
            column called ``wind_turbine_rotor_diameter`` may be
            included to specify a site-specific turbine diameter for
            each location). Columns that do not correspond to a config
            key may also be included, but they will be ignored. A
            DataFrame following the same guidelines as the CSV input
            (or a dictionary that can be used to initialize such a
            DataFrame) may be used for this input as well.
        sam_files : dict | str
            A dictionary mapping SAM input configuration ID(s) to SAM
            configuration(s). Keys are the SAM config ID(s) which
            correspond to the ``config`` column in the project points
            CSV. Values for each key are either a path to a
            corresponding SAM config file or a full dictionary
            of SAM config inputs. For example::

                sam_files = {
                    "default": "/path/to/default/sam.json",
                    "onshore": "/path/to/onshore/sam_config.yaml",
                    "offshore": {
                        "sam_key_1": "sam_value_1",
                        "sam_key_2": "sam_value_2",
                        ...
                    },
                    ...
                }

            This input can also be a string pointing to a single SAM
            config file. In this case, the ``config`` column of the
            CSV points input should be set to ``None`` or left out
            completely. See the documentation for the ``reV`` SAM class
            (e.g. :class:`reV.SAM.generation.WindPower`,
            :class:`reV.SAM.generation.PvWattsv8`,
            :class:`reV.SAM.generation.Geothermal`, etc.) for
            documentation on the allowed and/or required SAM config file
            inputs.
        cf_file : str
            Path to reV output generation file containing a capacity
            factor output.

            .. Note:: If executing ``reV`` from the command line, this
              path can contain brackets ``{}`` that will be filled in
              by the `analysis_years` input. Alternatively, this input
              can be set to ``"PIPELINE"`` to parse the output of the
              previous step (``reV`` generation) and use it as input to
              this call. However, note that duplicate executions of
              ``reV`` generation within the pipeline may invalidate this
              parsing, meaning the `cf_file` input will have to be
              specified manually.

        site_data : str | pd.DataFrame, optional
            Site-specific input data for SAM calculation. If this input
            is a string, it should be a path that points to a CSV file.
            Otherwise, this input should be a DataFrame with
            pre-extracted site data. Rows in this table should match
            the input sites via a ``gid`` column. The rest of the
            columns should match configuration input keys that will take
            site-specific values. Note that some or all site-specific
            inputs can be specified via the `project_points` input
            table instead. If ``None``, no site-specific data is
            considered. By default, ``None``.
        output_request : list | tuple, optional
            List of output variables requested from SAM. Can be any
            of the parameters in the "Outputs" group of the PySAM module
            (e.g. :py:class:`PySAM.Windpower.Windpower.Outputs`,
            :py:class:`PySAM.Pvwattsv8.Pvwattsv8.Outputs`,
            :py:class:`PySAM.Geothermal.Geothermal.Outputs`, etc.) being
            executed. This list can also include a select number of SAM
            config/resource parameters to include in the output:
            any key in any of the
            `output attribute JSON files <https://tinyurl.com/4bmrpe3j/>`_
            may be requested. Time-series profiles requested via this
            input are output in UTC. By default, ``('lcoe_fcr',)``.
        sites_per_worker : int, optional
            Number of sites to run in series on a worker. ``None``
            defaults to the resource file chunk size.
            By default, ``None``.
        memory_utilization_limit : float, optional
            Memory utilization limit (fractional). Must be a value
            between 0 and 1. This input sets how many site results will
            be stored in-memory at any given time before flushing to
            disk. By default, ``0.4``.
        append : bool
            Option to append econ datasets to source `cf_file`.
            By default, ``False``.
        """

        # get a points control instance
        pc = self.get_pc(
            points=project_points,
            points_range=None,
            sam_configs=sam_files,
            cf_file=cf_file,
            sites_per_worker=sites_per_worker,
            append=append,
        )

        super().__init__(
            pc,
            output_request,
            site_data=site_data,
            memory_utilization_limit=memory_utilization_limit,
        )

        self._cf_file = cf_file
        self._append = append
        self._run_attrs["cf_file"] = cf_file
        self._run_attrs["sam_module"] = self._sam_module.MODULE

    @property
    def cf_file(self):
        """Get the capacity factor output filename and path.

        Returns
        -------
        cf_file : str
            reV generation capacity factor output file with path.
        """
        return self._cf_file

    @property
    def meta(self):
        """Get meta data from the source capacity factors file.

        Returns
        -------
        _meta : pd.DataFrame
            Meta data from capacity factor outputs file.
        """
        if self._meta is None and self.cf_file is not None:
            with Outputs(self.cf_file) as cfh:
                # only take meta that belongs to this project's site list
                self._meta = cfh.meta[
                    cfh.meta[ResourceMetaField.GID].isin(
                        self.points_control.sites)]

            if ("offshore" in self._meta and self._meta["offshore"].sum() > 1):
                w = ('Found offshore sites in econ meta data. '
                     'This functionality has been deprecated. '
                     'Please run the reV offshore module to '
                     'calculate offshore wind lcoe.')
                warn(w, OffshoreWindInputWarning)
                logger.warning(w)

        elif self._meta is None and self.cf_file is None:
            self._meta = pd.DataFrame(
                {ResourceMetaField.GID: self.points_control.sites})

        return self._meta

    @property
    def time_index(self):
        """Get the generation resource time index data."""
        if self._time_index is None and self.cf_file is not None:
            with Outputs(self.cf_file) as cfh:
                if "time_index" in cfh.datasets:
                    self._time_index = cfh.time_index

        return self._time_index

    @staticmethod
    def _econ_append_pc(pp, cf_file, sites_per_worker=None):
        """
        Generate ProjectControls for econ append

        Parameters
        ----------
        pp : reV.config.project_points.ProjectPoints
            ProjectPoints to adjust gids for
        cf_file : str
            reV generation capacity factor output file with path.
        sites_per_worker : int
            Number of sites to run in series on a worker. None defaults to the
            resource file chunk size.

        Returns
        -------
        pc : reV.config.project_points.PointsControl
            PointsControl object instance.
        """
        multi_h5_res, hsds = check_res_file(cf_file)
        if multi_h5_res:
            res_cls = MultiFileResource
            res_kwargs = {}
        else:
            res_cls = Resource
            res_kwargs = {"hsds": hsds}

        with res_cls(cf_file, **res_kwargs) as f:
            gid0 = f.meta[ResourceMetaField.GID].values[0]
            gid1 = f.meta[ResourceMetaField.GID].values[-1]

        i0 = pp.index(gid0)
        i1 = pp.index(gid1) + 1
        pc = PointsControl.split(i0, i1, pp, sites_per_split=sites_per_worker)

        return pc

    @classmethod
    def get_pc(
        cls,
        points,
        points_range,
        sam_configs,
        cf_file,
        sites_per_worker=None,
        append=False,
    ):
        """
        Get a PointsControl instance.

        Parameters
        ----------
        points : slice | list | str | reV.config.project_points.PointsControl
            Slice specifying project points, or string pointing to a project
            points csv, or a fully instantiated PointsControl object.
        points_range : list | None
            Optional two-entry list specifying the index range of the sites to
            analyze. To be taken from the reV.config.PointsControl.split_range
            property.
        sam_configs : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s). Keys are the SAM
            config ID(s) which map to the config column in the project points
            CSV. Values are either a JSON SAM config file or dictionary of SAM
            config inputs. Can also be a single config file path or a
            pre loaded SAMConfig object.
        cf_file : str
            reV generation capacity factor output file with path.
        sites_per_worker : int
            Number of sites to run in series on a worker. None defaults to the
            resource file chunk size.
        append : bool
            Flag to append econ datasets to source cf_file. This has priority
            over the out_fpath input.

        Returns
        -------
        pc : reV.config.project_points.PointsControl
            PointsControl object instance.
        """
        pc = super().get_pc(
            points,
            points_range,
            sam_configs,
            ModuleName.ECON,
            sites_per_worker=sites_per_worker,
            res_file=cf_file,
        )

        if append:
            pc = cls._econ_append_pc(
                pc.project_points, cf_file, sites_per_worker=sites_per_worker
            )

        return pc

    @staticmethod
    def _run_single_worker(pc, econ_fun, output_request, **kwargs):
        """Run the SAM econ calculation.

        Parameters
        ----------
        pc : reV.config.project_points.PointsControl
            Iterable points control object from reV config module.
            Must have project_points with df property with all relevant
            site-specific inputs and a `SiteDataField.GID` column.
            By passing site-specific inputs in this dataframe, which
            was split using points_control, only the data relevant to
            the current sites is passed.
        econ_fun : method
            reV_run() method from one of the econ modules (SingleOwner,
            SAM_LCOE, WindBos).
        output_request : str | list | tuple
            Economic output variable(s) requested from SAM.
        kwargs : dict
            Additional input parameters for the SAM run module.

        Returns
        -------
        out : dict
            Output dictionary from the SAM reV_run function. Data is scaled
            within this function to the datatype specified in Econ.OUT_ATTRS.
        """

        # make sure output request is a list
        if isinstance(output_request, str):
            output_request = [output_request]

        # Extract the site df from the project points df.
        site_df = pc.project_points.df
        site_df = site_df.set_index(ResourceMetaField.GID, drop=True)

        # SAM execute econ analysis based on output request
        try:
            out = econ_fun(
                pc, site_df, output_request=output_request, **kwargs
            )
        except Exception as e:
            out = {}
            logger.exception("Worker failed for PC: {}".format(pc))
            raise e

        return out

    def _parse_output_request(self, req):
        """Set the output variables requested from generation.

        Parameters
        ----------
        req : str| list | tuple
            Output variables requested from SAM.

        Returns
        -------
        output_request : list
            Output variables requested from SAM.
        """

        output_request = super()._parse_output_request(req)

        for request in output_request:
            if request not in self.OUT_ATTRS:
                msg = (
                    'User output request "{}" not recognized. '
                    "Will attempt to extract from PySAM.".format(request)
                )
                logger.debug(msg)

        modules = [self.OPTIONS[request] for request in output_request
                   if request in self.OPTIONS]

        if not any(modules):
            msg = (
                "None of the user output requests were recognized. "
                "Cannot run reV econ. "
                "At least one of the following must be requested: {}".format(
                    list(self.OPTIONS.keys())
                )
            )
            logger.exception(msg)
            raise ExecutionError(msg)

        b1 = [m == modules[0] for m in modules]
        b2 = np.array([m == WindBos for m in modules])
        b3 = np.array([m == SingleOwner for m in modules])

        if all(b1):
            self._sam_module = modules[0]
            self._fun = modules[0].reV_run
        elif all(b2 | b3):
            self._sam_module = SingleOwner
            self._fun = SingleOwner.reV_run
        else:
            msg = (
                "Econ outputs requested from different SAM modules not "
                "currently supported. Output request variables require "
                "SAM methods: {}".format(modules)
            )
            raise ValueError(msg)

        return list(set(output_request))

    def _get_data_shape(self, dset, n_sites):
        """Get the output array shape based on OUT_ATTRS or PySAM.Outputs.

        This Econ get data shape method will also first check for the dset in
        the site_data table. If not found in site_data, the dataset will be
        looked for in OUT_ATTRS and PySAM.Outputs as it would for Generation.

        Parameters
        ----------
        dset : str
            Variable name to get shape for.
        n_sites : int
            Number of sites for this data shape.

        Returns
        -------
        shape : tuple
            1D or 2D shape tuple for dset.
        """

        if dset in self.site_data:
            data_shape = (n_sites,)
            data = self.site_data[dset].values[0]

            if isinstance(data, (list, tuple, np.ndarray, str)):
                msg = (
                    "Cannot pass through non-scalar site_data "
                    'input key "{}" as an output_request!'.format(dset)
                )
                logger.error(msg)
                raise ExecutionError(msg)

        else:
            data_shape = super()._get_data_shape(dset, n_sites)

        return data_shape

    def run(self, out_fpath=None, max_workers=1, timeout=1800, pool_size=None):
        """Execute a parallel reV econ run with smart data flushing.

        Parameters
        ----------
        out_fpath : str, optional
            Path to output file. If this class was initialized with
            ``append=True``, this input has no effect. If ``None``, no
            output file will be written. If the filepath is specified
            but the module name (econ) and/or resource data year is not
            included, the module name and/or resource data year will get
            added to the output file name. By default, ``None``.
        max_workers : int, optional
            Number of local workers to run on. By default, ``1``.
        timeout : int, optional
            Number of seconds to wait for parallel run iteration to
            complete before returning zeros. By default, ``1800``
            seconds.
        pool_size : int, optional
            Number of futures to submit to a single process pool for
            parallel futures. If ``None``, the pool size is set to
            ``os.cpu_count() * 2``. By default, ``None``.

        Returns
        -------
        str | None
            Path to output HDF5 file, or ``None`` if results were not
            written to disk.
        """
        if pool_size is None:
            pool_size = os.cpu_count() * 2

        # initialize output file or append econ data to gen file
        if self._append:
            self._out_fpath = self._cf_file
        else:
            self._init_fpath(out_fpath, ModuleName.ECON)

        self._init_h5(mode="a" if self._append else "w")
        self._init_out_arrays()

        diff = list(set(self.points_control.sites)
                    - set(self.meta[ResourceMetaField.GID].values))
        if diff:
            raise Exception(
                "The following analysis sites were requested "
                "through project points for econ but are not "
                'found in the CF file ("{}"): {}'.format(self.cf_file, diff)
            )

        # make a kwarg dict
        kwargs = {
            "output_request": self.output_request,
            "cf_file": self.cf_file,
            "year": self.year,
        }

        logger.info(
            "Running econ with smart data flushing " "for: {}".format(
                self.points_control
            )
        )
        logger.debug(
            'The following project points were specified: "{}"'.format(
                self.project_points
            )
        )
        logger.debug(
            "The following SAM configs are available to this run:\n{}".format(
                pprint.pformat(self.sam_configs, indent=4)
            )
        )
        logger.debug(
            "The SAM output variables have been requested:\n{}".format(
                self.output_request
            )
        )

        try:
            kwargs["econ_fun"] = self._fun
            if max_workers == 1:
                logger.debug(
                    "Running serial econ for: {}".format(self.points_control)
                )
                for i, pc_sub in enumerate(self.points_control):
                    self.out = self._run_single_worker(pc_sub, **kwargs)
                    logger.info(
                        "Finished reV econ serial compute for: {} "
                        "(iteration {} out of {})".format(
                            pc_sub, i + 1, len(self.points_control)
                        )
                    )
                self.flush()
            else:
                logger.debug(
                    "Running parallel econ for: {}".format(self.points_control)
                )
                self._parallel_run(
                    max_workers=max_workers,
                    pool_size=pool_size,
                    timeout=timeout,
                    **kwargs,
                )

        except Exception as e:
            logger.exception("SmartParallelJob.execute() failed for econ.")
            raise e

        return self._out_fpath



# -*- coding: utf-8 -*-
"""reV utilities."""
from enum import Enum

import PySAM
from rex.utilities.loggers import log_versions as rex_log_versions

from reV.version import __version__


class FieldEnum(str, Enum):
    """Base Field enum with some mapping methods."""

    @classmethod
    def map_to(cls, other):
        """Return a rename map from this enum to another.

        Mapping is performed on matching enum names. In other words, if
        both enums have a `ONE` attribute, this will be mapped from one
        enum to another.

        Parameters
        ----------
        other : :class:`Enum`
            ``Enum`` subclass with ``__members__`` attribute.

        Returns
        -------
        dict
            Dictionary mapping matching values from one enum to another.

        Examples
        --------
        >>> class Test1(FieldEnum):
        >>>     ONE = "one_x"
        >>>     TWO = "two"
        >>>
        >>> class Test2(Enum):
        >>>     ONE = "one_y"
        >>>     THREE = "three"
        >>>
        >>> Test1.map_to(Test2)
        {<Test1.ONE: 'one_x'>: <Test2.ONE: 'one_y'>}
        """
        return {
            cls[mem]: other[mem]
            for mem in cls.__members__
            if mem in other.__members__
        }

    @classmethod
    def map_from(cls, other):
        """Map from a dictionary of name / member pairs to this enum.

        Parameters
        ----------
        other : dict
            Dictionary mapping key values (typically old aliases) to
            enum values. For example, ``{'sc_gid': 'SC_GID'}`` would
            return a dictionary that maps ``'sc_gid'`` to the ``SC_GID``
            member of this enum.

        Returns
        -------
        dict
            Mapping of input dictionary keys to member values of this
            enum.

        Examples
        --------
        >>> class Test(FieldEnum):
        >>>     ONE = "one_x"
        >>>     TWO = "two_y"
        >>>
        >>> Test.map_from({1: "ONE", 2: "TWO"})
        {1: <Test.ONE: 'one_x'>, 2: <Test.TWO: 'two_y'>}
        """
        return {name: cls[mem] for name, mem in other.items()}

    def __str__(self):
        return self.value

    def __format__(self, format_spec):
        return str.__format__(self.value, format_spec)


class SiteDataField(FieldEnum):
    """An enumerated map to site data column names."""

    GID = "gid"
    CONFIG = "config"


class ResourceMetaField(FieldEnum):
    """An enumerated map to resource meta column names.

    Each output name should match the name of a key the resource file
    meta table.
    """

    GID = "gid"
    LATITUDE = "latitude"
    LONGITUDE = "longitude"
    ELEVATION = "elevation"
    TIMEZONE = "timezone"
    COUNTY = "county"
    STATE = "state"
    COUNTRY = "country"
    OFFSHORE = "offshore"


class SupplyCurveField(FieldEnum):
    """An enumerated map to supply curve summary/meta keys.

    This is a collection of known supply curve fields that reV outputs
    across aggregation, supply curve, and bespoke outputs.

    Not all of these columns are guaranteed in every supply-curve like
    output (e.g. "convex_hull_area" is a bespoke-only output).
    """

    SC_GID = "sc_gid"
    LATITUDE = "latitude"
    LONGITUDE = "longitude"
    COUNTRY = "country"
    STATE = "state"
    COUNTY = "county"
    ELEVATION = "elevation_m"
    TIMEZONE = "timezone"
    SC_POINT_GID = "sc_point_gid"
    SC_ROW_IND = "sc_row_ind"
    SC_COL_IND = "sc_col_ind"
    SOURCE_GIDS = "source_gids"
    RES_GIDS = "res_gids"
    GEN_GIDS = "gen_gids"
    GID_COUNTS = "gid_counts"
    N_GIDS = "n_gids"
    MEAN_RES = "resource"
    MEAN_CF_AC = "capacity_factor_ac"
    MEAN_CF_DC = "capacity_factor_dc"
    MEAN_LCOE = "lcoe_site_usd_per_mwh"
    CAPACITY_AC_MW = "capacity_ac_mw"
    CAPACITY_DC_MW = "capacity_dc_mw"
    OFFSHORE = "offshore"
    AREA_SQ_KM = "area_developable_sq_km"
    MEAN_FRICTION = "friction_site"
    MEAN_LCOE_FRICTION = "lcoe_friction_usd_per_mwh"
    RAW_LCOE = "lcoe_raw_usd_per_mwh"
    EOS_MULT = "multiplier_cc_eos"
    REG_MULT = "multiplier_cc_regional"
    SC_POINT_ANNUAL_ENERGY_MW = "annual_energy_site_mwh"
    COST_BASE_OCC_USD_PER_AC_MW = "cost_base_occ_usd_per_ac_mw"
    COST_SITE_OCC_USD_PER_AC_MW = "cost_site_occ_usd_per_ac_mw"
    COST_BASE_FOC_USD_PER_AC_MW = "cost_base_foc_usd_per_ac_mw"
    COST_SITE_FOC_USD_PER_AC_MW = "cost_site_foc_usd_per_ac_mw"
    COST_BASE_VOC_USD_PER_AC_MW = "cost_base_voc_usd_per_ac_mw"
    COST_SITE_VOC_USD_PER_AC_MW = "cost_site_voc_usd_per_ac_mw"
    FIXED_CHARGE_RATE = "fixed_charge_rate"

    # Bespoke outputs
    POSSIBLE_X_COORDS = "possible_x_coords"
    POSSIBLE_Y_COORDS = "possible_y_coords"
    TURBINE_X_COORDS = "turbine_x_coords"
    TURBINE_Y_COORDS = "turbine_y_coords"
    N_TURBINES = "n_turbines"
    INCLUDED_AREA = "area_included_sq_km"
    INCLUDED_AREA_CAPACITY_DENSITY = (
        "capacity_density_included_area_mw_per_km2"
    )
    CONVEX_HULL_AREA = "area_convex_hull_sq_km"
    CONVEX_HULL_CAPACITY_DENSITY = "capacity_density_convex_hull_mw_per_km2"
    FULL_CELL_CAPACITY_DENSITY = "capacity_density_full_cell_mw_per_km2"
    BESPOKE_AEP = "optimized_plant_aep"
    BESPOKE_OBJECTIVE = "optimized_plant_objective"
    BESPOKE_CAPITAL_COST = "optimized_plant_capital_cost"
    BESPOKE_FIXED_OPERATING_COST = "optimized_plant_fixed_operating_cost"
    BESPOKE_VARIABLE_OPERATING_COST = "optimized_plant_variable_operating_cost"
    BESPOKE_BALANCE_OF_SYSTEM_COST = "optimized_plant_balance_of_system_cost"

    # Transmission outputs
    TRANS_GID = "trans_gid"
    TRANS_TYPE = "trans_type"
    TOTAL_LCOE_FRICTION = "lcoe_total_friction_usd_per_mwh"
    TRANS_CAPACITY = "trans_capacity"
    DIST_SPUR_KM = "dist_spur_km"
    DIST_EXPORT_KM = "dist_export_km"
    REINFORCEMENT_DIST_KM = "dist_reinforcement_km"
    TIE_LINE_COST_PER_MW = "cost_spur_usd_per_mw"
    CONNECTION_COST_PER_MW = "cost_poi_usd_per_mw"
    EXPORT_COST_PER_MW = "cost_export_usd_per_mw"
    REINFORCEMENT_COST_PER_MW = "cost_reinforcement_usd_per_mw"
    TOTAL_TRANS_CAP_COST_PER_MW = "cost_total_trans_usd_per_mw"
    LCOT = "lcot_usd_per_mwh"
    TOTAL_LCOE = "lcoe_all_in_usd_per_mwh"
    N_PARALLEL_TRANS = "count_num_parallel_trans"
    POI_LAT = "latitude_poi"
    POI_LON = "longitude_poi"
    REINFORCEMENT_POI_LAT = "latitude_reinforcement_poi"
    REINFORCEMENT_POI_LON = "longitude_reinforcement_poi"

    @classmethod
    def map_from_legacy(cls):
        """Map of legacy names to current values.

        Returns
        -------
        dict
            Dictionary that maps legacy supply curve column names to
            members of this enum.
        """
        legacy_map = {}
        for current_field, old_field in cls.map_to(_LegacySCAliases).items():
            aliases = old_field.value
            if isinstance(aliases, str):
                aliases = [aliases]
            legacy_map.update({alias: current_field for alias in aliases})

        return legacy_map


class _LegacySCAliases(Enum):
    """Legacy supply curve column names.

    Enum values can be either a single string or an iterable of string
    values where each string value represents a previously known alias.
    """

    ELEVATION = "elevation"
    MEAN_RES = "mean_res"
    MEAN_CF_AC = "mean_cf"
    MEAN_LCOE = "mean_lcoe"
    CAPACITY_AC_MW = "capacity"
    AREA_SQ_KM = "area_sq_km"
    MEAN_FRICTION = "mean_friction"
    MEAN_LCOE_FRICTION = "mean_lcoe_friction"
    RAW_LCOE = "raw_lcoe"
    TRANS_TYPE = "category"
    TRANS_CAPACITY = "avail_cap"
    DIST_SPUR_KM = "dist_km"
    REINFORCEMENT_DIST_KM = "reinforcement_dist_km"
    TIE_LINE_COST_PER_MW = "tie_line_cost_per_mw"
    CONNECTION_COST_PER_MW = "connection_cost_per_mw"
    REINFORCEMENT_COST_PER_MW = "reinforcement_cost_per_mw"
    TOTAL_TRANS_CAP_COST_PER_MW = "trans_cap_cost_per_mw"
    LCOT = "lcot"
    TOTAL_LCOE = "total_lcoe"
    TOTAL_LCOE_FRICTION = "total_lcoe_friction"
    N_PARALLEL_TRANS = "n_parallel_trans"
    EOS_MULT = "eos_mult", "capital_cost_multiplier"
    REG_MULT = "reg_mult"
    SC_POINT_ANNUAL_ENERGY_MW = "sc_point_annual_energy"
    POI_LAT = "poi_lat"
    POI_LON = "poi_lon"
    REINFORCEMENT_POI_LAT = "reinforcement_poi_lat"
    REINFORCEMENT_POI_LON = "reinforcement_poi_lon"
    BESPOKE_AEP = "bespoke_aep"
    BESPOKE_OBJECTIVE = "bespoke_objective"
    BESPOKE_CAPITAL_COST = "bespoke_capital_cost"
    BESPOKE_FIXED_OPERATING_COST = "bespoke_fixed_operating_cost"
    BESPOKE_VARIABLE_OPERATING_COST = "bespoke_variable_operating_cost"
    BESPOKE_BALANCE_OF_SYSTEM_COST = "bespoke_balance_of_system_cost"
    INCLUDED_AREA = "included_area"
    INCLUDED_AREA_CAPACITY_DENSITY = "included_area_capacity_density"
    CONVEX_HULL_AREA = "convex_hull_area"
    CONVEX_HULL_CAPACITY_DENSITY = "convex_hull_capacity_density"
    FULL_CELL_CAPACITY_DENSITY = "full_cell_capacity_density"


class ModuleName(str, Enum):
    """A collection of the module names available in reV.

    Each module name should match the name of the click command
    that will be used to invoke its respective cli. As of 3/1/2022,
    this means that all commands are lowercase with underscores
    replaced by dashes.

    Reference
    ---------
    See this line in the click source code to get the most up-to-date
    click name conversions: https://tinyurl.com/4rehbsvf
    """

    BESPOKE = "bespoke"
    COLLECT = "collect"
    ECON = "econ"
    GENERATION = "generation"
    HYBRIDS = "hybrids"
    MULTI_YEAR = "multi-year"
    NRWAL = "nrwal"
    QA_QC = "qa-qc"
    REP_PROFILES = "rep-profiles"
    SUPPLY_CURVE = "supply-curve"
    SUPPLY_CURVE_AGGREGATION = "supply-curve-aggregation"

    def __str__(self):
        return self.value

    def __format__(self, format_spec):
        return str.__format__(self.value, format_spec)

    @classmethod
    def all_names(cls):
        """All module names.

        Returns
        -------
        set
            The set of all module name strings.
        """
        # pylint: disable=no-member
        return {v.value for v in cls.__members__.values()}


def log_versions(logger):
    """Log package versions:
    - rex and reV to info
    - h5py, numpy, pandas, scipy, and PySAM to debug

    Parameters
    ----------
    logger : logging.Logger
        Logger object to log memory message to.
    """
    logger.info("Running with reV version {}".format(__version__))
    rex_log_versions(logger)
    logger.debug("- PySAM version {}".format(PySAM.__version__))



# -*- coding: utf-8 -*-
"""
Econ CLI utility functions.
"""
import pprint
import os
import logging

from rex.utilities.utilities import parse_year
from gaps.pipeline import parse_previous_status
from gaps.cli import as_click_command, CLICommandFromClass

from reV.econ.econ import Econ
from reV.utilities import ModuleName
from reV.utilities.cli_functions import format_analysis_years, init_cli_logging
from reV.utilities.exceptions import ConfigError


logger = logging.getLogger(__name__)


def _preprocessor(config, out_dir, job_name, log_directory, verbose,
                  analysis_years=None):
    """Preprocess econ config user input.

    Parameters
    ----------
    config : dict
        User configuration file input as (nested) dict.
    out_dir : str
        Path to output file directory.
    job_name : str
        Name of ``reV`` job being run.
    log_directory : str
        Path to log output directory.
    verbose : bool
        Flag to signal ``DEBUG`` verbosity (``verbose=True``).
    analysis_years : int | list, optional
        A single year or list of years to perform analysis for. These
        years will be used to fill in any brackets ``{}`` in the
        ``resource_file`` input. If ``None``, the ``resource_file``
        input is assumed to be the full path to the single resource
        file to be processed.  By default, ``None``.

    Returns
    -------
    dict
        Updated config file.
    """
    init_cli_logging(job_name, log_directory, verbose)
    analysis_years = format_analysis_years(analysis_years)
    config["cf_file"] = _parse_cf_files(config["cf_file"], analysis_years,
                                        out_dir)

    _log_econ_cli_inputs(config)
    return config


def _parse_cf_files(cf_file, analysis_years, out_dir):
    """Get the capacity factor files (reV generation output data). """

    # get base filename, may have {} for year format
    if '{}' in cf_file:
        # need to make list of res files for each year
        cf_files = [cf_file.format(year) for year in analysis_years]
    elif 'PIPELINE' in cf_file:
        cf_files = parse_previous_status(out_dir, command=str(ModuleName.ECON))
    else:
        # only one resource file request, still put in list
        cf_files = [cf_file]

    for f in cf_files:
        # ignore files that are to be specified using pipeline utils
        if 'PIPELINE' not in os.path.basename(f):
            if not os.path.exists(f):
                raise IOError('File does not exist: {}'.format(f))

    # check year/cf_file matching if not a pipeline input
    if 'PIPELINE' not in cf_file:
        if len(cf_files) != len(analysis_years):
            raise ConfigError('The number of cf files does not match '
                              'the number of analysis years!'
                              '\n\tCF files: \n\t\t{}'
                              '\n\tYears: \n\t\t{}'
                              .format(cf_files, analysis_years))
        for year in analysis_years:
            if str(year) not in str(cf_files):
                raise ConfigError('Could not find year {} in cf '
                                  'files: {}'.format(year, cf_files))

    return [fn for fn in cf_files if parse_year(fn) in analysis_years]


def _log_econ_cli_inputs(config):
    """Log initial econ CLI inputs"""

    logger.info('The following project points were specified: "{}"'
                .format(config.get('project_points', None)))
    logger.info('The following SAM configs are available to this run:\n{}'
                .format(pprint.pformat(config.get('sam_files', None),
                                       indent=4)))
    logger.debug('Submitting jobs for the following cf_files: {}'
                 .format(config.get("cf_file")))


econ_command = CLICommandFromClass(Econ, method="run",
                                   name=str(ModuleName.ECON),
                                   add_collect=False,
                                   split_keys=["project_points", "cf_file"],
                                   config_preprocessor=_preprocessor)
main = as_click_command(econ_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Econ CLI.')
        raise



# -*- coding: utf-8 -*-
"""
Bespoke wind plant optimization CLI utility functions.
"""
import pprint
import logging

from reV.bespoke.bespoke import BespokeWindPlants
from reV.utilities import ModuleName
from reV.utilities.cli_functions import init_cli_logging
from gaps.cli import as_click_command, CLICommandFromClass


logger = logging.getLogger(__name__)


def _preprocessor(config, job_name, log_directory, verbose):
    """Preprocess bespoke config user input.

    Parameters
    ----------
    config : dict
        User configuration file input as (nested) dict.
    job_name : str
        Name of ``reV`` job being run.
    log_directory : str
        Path to log output directory.
    verbose : bool
        Flag to signal ``DEBUG`` verbosity (``verbose=True``).

    Returns
    -------
    dict
        Updated config file.
    """
    init_cli_logging(job_name, log_directory, verbose)
    if isinstance(config["sam_files"], str):
        config["sam_files"] = {'default': config["sam_files"]}

    _log_bespoke_cli_inputs(config)
    return config


def _log_bespoke_cli_inputs(config):
    """Log initial bespoke CLI inputs"""

    logger.info('Source resource file: "{}"'.format(config.get("res_fpath")))
    logger.info('Source exclusion file: "{}"'.format(config.get("excl_fpath")))
    logger.info('Bespoke optimization objective function: "{}"'
                .format(config.get("objective_function")))
    logger.info('Bespoke capital cost function: "{}"'
                .format(config.get("capital_cost_function")))
    logger.info('Bespoke fixed operating cost function: "{}"'
                .format(config.get("fixed_operating_cost_function")))
    logger.info('Bespoke variable operating cost function: "{}"'
                .format(config.get("variable_operating_cost_function")))
    logger.info('Bespoke balance of system cost function: "{}"'
                .format(config.get("balance_of_system_cost_function")))
    logger.info('Bespoke wake loss multiplier: "{}"'
                .format(config.get("wake_loss_multiplier", 1)))
    logger.info('The following project points were specified: "{}"'
                .format(config.get('project_points', None)))
    logger.info('The following SAM configs are available to this run:\n{}'
                .format(pprint.pformat(config.get('sam_files'), indent=4)))
    logger.debug('The following exclusion dictionary was specified\n{}'
                 .format(pprint.pformat(config.get('excl_dict'), indent=4)))


bespoke_command = CLICommandFromClass(BespokeWindPlants, method="run",
                                      name=str(ModuleName.BESPOKE),
                                      add_collect=False,
                                      split_keys=["project_points"],
                                      config_preprocessor=_preprocessor)
main = as_click_command(bespoke_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Bespoke CLI.')
        raise



# -*- coding: utf-8 -*-
"""
reV-SAM interface module
"""
from .econ import LCOE, SingleOwner
from .generation import (PvWattsv5, PvWattsv7, PvWattsv8, PvSamv1,
                         TcsMoltenSalt, WindPower)
from .windbos import WindBos
